/*
 *  Copyright (C) 2012 Khaldoon Ghanem
 *  See the file license.txt for copying permission.
 */

#include <stdio.h>
#include <stdlib.h>


#if defined(WIN32)
	#define GLEW_STATIC
	#include <GL/glew.h>
	#include <windows.h>
#else
	#include <GL/glew.h>
#endif


#if defined(__APPLE__) || defined(MACOSX)
	#include <OpenGL/gl.h>
	#include <OpenGL/glu.h>
	#include <GLUT/glut.h>
#else
	#include <GL/glut.h>
	#include <GL/gl.h>
	#include <GL/glu.h>
#endif


#include <math.h>
#include <string.h>

#include "tables.c"

float* verts;
float* normals;
GLuint* triangs;
int totalVerts = 0;
int totalTriangs = 0;

float* vertsR;
float* normalsR;
GLuint* triangsR;
int totalVertsR = 0;
int totalTriangsR = 0;

float* vertsB;
float* normalsB;
GLuint* triangsB;
int totalVertsB;
int totalTriangsB;

float pi;
float r, theta, phi;
int smooth = 1;
void init(void)
{
    GLfloat mat_ambient[] = { 1.0, 1.0, 1.0, 1.0 };
    GLfloat mat_specular[] = { 1.0, 1.0, 1.0, 1.0 };
    GLfloat mat_shininess[] = { 100.0 };
    GLfloat white_light[] = { 1.0, 1.0, 1.0, 1.0 };
    GLfloat lmodel_ambient[] = { 0.1, 0.1, 0.1, 1.0 };
    GLfloat light_position[] = { 0.0, 0.0, 1.0, 0.0 };
    
    glClearColor(0.0, 0.0, 0.0, 0.0);
    glShadeModel (GL_SMOOTH);

    glMaterialfv(GL_FRONT, GL_AMBIENT, mat_ambient);
    glMaterialfv(GL_FRONT, GL_SPECULAR, mat_specular);
    glMaterialfv(GL_FRONT, GL_SHININESS, mat_shininess);
    
    glMaterialfv(GL_BACK, GL_AMBIENT, mat_ambient);
    glMaterialfv(GL_BACK, GL_SPECULAR, mat_specular);
    glMaterialfv(GL_BACK, GL_SHININESS, mat_shininess);    

    glLightfv(GL_LIGHT0, GL_DIFFUSE, white_light);
    glLightfv(GL_LIGHT0, GL_SPECULAR, white_light);                    
    glLightModelfv(GL_LIGHT_MODEL_AMBIENT, lmodel_ambient);
    glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_TRUE);
    
    glLightfv(GL_LIGHT0, GL_POSITION, light_position);
    glEnable (GL_DEPTH_TEST);
    glEnable(GL_LIGHTING);
    glEnable(GL_LIGHT0);

}

void drawGrid(float spacing, float min, float max)
{
    glDisable(GL_LIGHTING);
    float coord;
    glLineWidth(1);
    glBegin (GL_LINES);
    //Draw X-Y Plane            
    glColor3f(0.2, 0.2, 0.2);
    coord = spacing;
    while (coord <= max)
    {
        glVertex3f(coord, min, 0);
        glVertex3f(coord, max, 0);
        glVertex3f(min, coord, 0);
        glVertex3f(max, coord, 0);
        coord += spacing;
    }
    coord = -spacing;
    while (coord >= min)
    {
        glVertex3f(coord, min, 0);
        glVertex3f(coord, max, 0);
        glVertex3f(min, coord, 0);
        glVertex3f(max, coord, 0);
        coord -= spacing;
    }
    glEnd();
    glLineWidth(3);
    glBegin(GL_LINES);
    //Draw Axes    

    glColor3f(0.3, 0, 0);
    glVertex3f(min, 0, 0);
    glVertex3f(0, 0, 0);
    glColor3f(1, 0, 0);
    glVertex3f(0, 0, 0);
    glVertex3f(max, 0, 0);

    glColor3f(0, 0.3, 0);
    glVertex3f(0, min, 0);
    glVertex3f(0, 0, 0);
    glColor3f(0, 1, 0);
    glVertex3f(0, 0, 0);
    glVertex3f(0, max, 0);

    glColor3f(0, 0, 0.3);
    glVertex3f(0, 0, min);
    glVertex3f(0, 0, 0);
    glColor3f(0, 0, 1);
    glVertex3f(0, 0, 0);
    glVertex3f(0, 0, max);

    glEnd();
    glEnable(GL_LIGHTING);
}

GLfloat mat_diffuseR[] = { 1.0, 0.0, 0.0, 1.0 };
GLfloat mat_diffuseB[] = { 0.0, 0.0, 1.0, 1.0 };
void display(void)
{
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    //
    glLoadIdentity();
    float x = r * sin(theta) * cos(phi);
    float y = r * sin(theta) * sin(phi);
    float z = r * cos(theta);

    gluLookAt(x, y, z, 0.0, 0.0, 0.0, 0.0f, 0.0f, 1.0f);

    //Draw grid

    drawGrid(1, -30, 30);
    
    
    glMaterialfv(GL_FRONT, GL_DIFFUSE, mat_diffuseR);
    glMaterialfv(GL_BACK, GL_DIFFUSE, mat_diffuseR);
    if(totalVertsR!=0)
    {
        glEnableClientState(GL_VERTEX_ARRAY);
        glEnableClientState(GL_NORMAL_ARRAY);
        
        
        glVertexPointer(3, GL_FLOAT, 0, vertsR);
        glNormalPointer(GL_FLOAT, 0, normalsR);
        glDrawElements(GL_TRIANGLES, totalTriangsR*3, GL_UNSIGNED_INT, triangsR);              
        
        glDisableClientState(GL_NORMAL_ARRAY);
        glDisableClientState(GL_VERTEX_ARRAY);
    }
    
    glMaterialfv(GL_FRONT, GL_DIFFUSE, mat_diffuseB);
    glMaterialfv(GL_BACK, GL_DIFFUSE, mat_diffuseB);
    if(totalVertsB!=0)
    {
        glEnableClientState(GL_VERTEX_ARRAY);
        glEnableClientState(GL_NORMAL_ARRAY);
        
        
        glVertexPointer(3, GL_FLOAT, 0, vertsB);
        glNormalPointer(GL_FLOAT, 0, normalsB);
        glDrawElements(GL_TRIANGLES, totalTriangsB*3, GL_UNSIGNED_INT, triangsB);               
        
        glDisableClientState(GL_NORMAL_ARRAY);
        glDisableClientState(GL_VERTEX_ARRAY);
    }    
    

    //glutSolidSphere(1.0, 256, 256);
    glutSwapBuffers();
}

int gw, gh;
void reshape(int w, int h)
{
    // Prevent a divide by zero, when window is too short
    // (you cant make a window of zero width).
    if (h == 0)
        h = 1;

    float ratio = 1.0 * w / h;

    // Reset the coordinate system before modifying
    glMatrixMode (GL_PROJECTION);
    glLoadIdentity();

    // Set the viewport to be the entire window
    glViewport(0, 0, w, h);

    // Set the correct perspective.
    gluPerspective(45, ratio, 0.5, 500);
    glMatrixMode (GL_MODELVIEW);
    gh = h;
    gw = w;
}

int fullscreen = 0;
int screen_pos_x;
int screen_pos_y;
int screen_w;
int screen_h;
float isoValue;
void generateIsosurface();

void generateRedBlueIso()
{
    /////Red/////
    if(totalVertsR!=0)
    {
        free(vertsR);
        free(triangsR);
        free(normalsR);
    }
    
    generateIsosurface();
    
    vertsR = verts;
    triangsR = triangs;
    normalsR = normals;
    totalTriangsR = totalTriangs;
    totalVertsR= totalVerts;
           
    /////Blue/////
    isoValue = -1 * isoValue;  
    
    if(totalVertsB!=0)
    {
        free(vertsB);
        free(triangsB);
        free(normalsB);
    }
    
    generateIsosurface();
    
    vertsB = verts;
    triangsB = triangs;
    normalsB = normals;
    totalTriangsB = totalTriangs;
    totalVertsB= totalVerts;        
    
    
    
    isoValue = -1 * isoValue;    
}
void processNormalKeys(unsigned char key, int x, int y)
{
    //Navigation
    if (key == 'w' || key == 'W')
    {
        if (theta - 0.1 > 0)
            theta -= 0.1;
    }
    if (key == 's' || key == 'S')
    {
        if (theta + 0.1 < pi)
            theta += 0.1;
    }

    if (key == 'a' || key == 'A')
    {
        phi -= 0.1;
    }
    if (key == 'd' || key == 'D')
    {
        phi += 0.1;
    }

    if (key == 'q' || key == 'Q')
    {
        r = r * 1.2;
    }

    if (key == 'e' || key == 'E')
    {
        r = r / 1.2;
    }

    if (key == 'p' || key == 'P')
    {
        printf("Camera Position:");
        printf("r = %lg, theta = %lg, phi= %lg ", r, theta, phi);
        float x = r * sin(theta) * cos(phi);
        float y = r * sin(theta) * sin(phi);
        float z = r * cos(theta);
        printf("(x = %lg, y = %lg, z= %lg).\n", x, y, z);
    }

    if (key == 'f' || key == 'F')
    {
        fullscreen = !fullscreen;
        if (fullscreen)
        {
            screen_pos_x = glutGet((GLenum) GLUT_WINDOW_X);
            screen_pos_y = glutGet((GLenum) GLUT_WINDOW_Y);
            screen_w = gw;
            screen_h = gh;
            glutFullScreen();
        }
        else
        {
            glutReshapeWindow(screen_w, screen_h);
            glutPositionWindow(screen_pos_x, screen_pos_y);
        }
    }
    if (key == 'i' || key == 'I')
    {
        printf("Isovlaue: %lg\n", isoValue);
    }
    if (key == 'u' || key == 'U')
    {
        if(smooth)
        {
            glShadeModel (GL_FLAT);
            smooth = 0;
        }
        else
        {
            glShadeModel (GL_SMOOTH);
            smooth = 1;
        } 
    }
    if (key == 'z' || key == 'Z')
    {
        
        isoValue /= 1.1;
        generateRedBlueIso();

    }
    if (key == 'x' || key == 'X')
    {
        isoValue *= 1.1;
        generateRedBlueIso();
    }
    if (key == 27)
        exit(0);
}


float mouse_x = 0;
float mouse_y = 0;
float old_theta = 0, old_phi = 0;
int pressed = 0;
void mouseButton(int button, int state, int x, int y)
{
    if (button == GLUT_LEFT_BUTTON)
    {

        // when the button is released
        if (state == GLUT_UP)
        {
            pressed = 0;
        }
        else
        { // state = GLUT_DOWN
            mouse_x = x;
            mouse_y = y;
            old_theta = theta;
            old_phi = phi;
            pressed = 1;
        }
    }
    // Wheel reports as button 3(scroll up) and button 4(scroll down)
    if ((button == 3 && state == GLUT_DOWN)
            || (button == 4 && state == GLUT_DOWN))
    {
        if (button == 4)
            r *= 1.2;
        else
            r /= 1.2;
//        if (button == 4)
//            isoValue *= 1.1;
//        else
//            isoValue /= 1.1;
//        generateRedBlueIso();
    }
}

void mouseMove(int x, int y)
{
    if (pressed)
    {
        theta = old_theta + (float) (mouse_y - y) / gh * pi;
        if (theta < 0)
            theta = 0.001;
        if (theta > pi)
            theta = pi - 0.001;
        phi = old_phi + (float) (mouse_x - x) / gw * pi;
    }
}

char *textFileRead(char *fn) {


	FILE *fp;
	char *content = NULL;

	int count=0;

	if (fn != NULL) {
		fp = fopen(fn,"rt");

		if (fp != NULL) {
      
      fseek(fp, 0, SEEK_END);
      count = ftell(fp);
      rewind(fp);

			if (count > 0) {
				content = (char *)malloc(sizeof(char) * (count+1));
				count = fread(content,sizeof(char),count,fp);
				content[count] = '\0';
			}
			fclose(fp);
		}
	}
	return content;
}


float*  field;
float e1[3];
float e2[3];
float e3[3];
float x0, y_0, z0;
int nx, ny, nz;

#define INDX(i, j, k) ((i)*ny*nz+(j)*nz+(k))
#define K(indx) ((indx)%nz)
#define J(indx) (((indx)/nz)%ny)
#define I(indx) ((indx)/(nz*ny))

float getX(int mp)
{
    return x0 + e1[0] * I(mp) + e2[0] * J(mp) + e3[0] * K(mp);
}

float getY(int mp)
{
    return y_0 + e1[1] * I(mp) + e2[1] * J(mp) + e3[1] * K(mp);
}

float getZ(int mp)
{
    return z0 + e1[2] * I(mp) + e2[2] * J(mp) + e3[2] * K(mp);
}

void loadField(char* fileName)
{
    printf("Loading field data from file: '%s' ...\n", fileName);
    FILE* fp = fopen(fileName, "r");   
    if(fp==0)
    {
        fprintf(stderr, "Can not open file!\n");
        exit(1);
    }

    char c = '0';
    while(c!='\n' && c!=EOF) c = fgetc(fp);
    c = '0';
    while(c!='\n' && c!=EOF) c = fgetc(fp);
    
    fscanf(fp, " %*d %g %g %g\n", &x0, &y_0 ,&z0);
    
    fscanf(fp, " %d %g %g %g\n", &nx, &e1[0], &e1[1], &e1[2]);
    fscanf(fp, " %d %g %g %g\n", &ny, &e2[0], &e2[1], &e2[2]);
    fscanf(fp, " %d %g %g %g\n", &nz, &e3[0], &e3[1], &e3[2]);
    field = (float*) malloc(sizeof(float) * nx * ny * nz);
    for(int i=0;i<nx;i++)
    {   
        for(int j=0;j<ny;j++)
        {
            for (int k=0;k<nz;k++)
            {
                fscanf(fp, " %g ", &field[INDX(i, j, k)]);
            }
        }
    }    
    fclose(fp);
}

#define EDGE_ID(i, j, k) ((i)*4+(j)*2+(k))
#define EDGE_DK(indx) ((indx)%2)
#define EDGE_DJ(indx) (((indx)/2)%2)
#define EDGE_DI(indx) ((indx)/4)

int totalMPs;
int* mpEdgeIndx;


int totalActive;
int* mpActive;
int* activeMPs;
int* mpActiveScanned;

int* mpVertsNum;
int* mpVertsBase;

float* voxNormals;

int* voxTriangsNum;
int* voxTriangsBase;


void generateIsosurface()
{
    printf("Generating Iso Surface:\n");
    printf("-----------------------\n");
//1-Generate Vertices & Normals:
//===============================
    printf("  1-Processing Vertices ...\n");
    
  //1a-Counting Vertices:
  //---------------------
    printf("    1a-Counting Vertices ...\n");
    
    totalMPs = nx*ny*nz;//(nx-1)*(ny-1)*(nz-1); 
    printf("      Total Mesh Points: %d\n", totalMPs);
    //using (nx-1)*(ny-1)*(nz-1) can save some memory but need changes to INDX    
    mpEdgeIndx = (int*) malloc(sizeof(int)* totalMPs);    
    mpVertsNum = (int*) malloc(sizeof(int)* totalMPs);
    mpActive = (int*) malloc(sizeof(int)* totalMPs);
    //We use a mapping between mesh points and mesh edges
    //In marching cubes it is  1 --> 3 mapping.
    //In marching tetrahedrons it is 1 --> 7 mapping.
    
    //Each mesh point (i,j,k) is associated with seven mesh edges:    
    //0:(i,j,k)-(i,   j,   k) Igonre this. Just listed to complete the pattern.
    //1: (i,j,k)-(i,   j,   k+1)
    //2: (i,j,k)-(i,   j+1, k)
    //3: (i,j,k)-(i,   j+1, k+1)
    //4: (i,j,k)-(i+1, j,   k)
    //5: (i,j,k)-(i+1, j,   k+1)
    //6: (i,j,k)-(i+1, j+1, k)
    //7: (i,j,k)-(i+1, j+1, k+1)
    //Edge number can be computed using di*4+dj*2+dk
    //where di, dj, dk is the displacments of the other end (0 or 1).
    
    //For each mesh point, we see which of the associated edges are crossed 
    //by the isosurface.
    
    //Bit number 'b' of mpEdgeIndx[mp] will indicate whether edge number 'b'
    //associated with mesh point 'mp' is crossed or not.
    
    //mpVertsNum[mp] is the number of isosurface vertices associated with
    //mesh point 'mp' (it is the same as the number of crossed edges)
   
    for(int mp=0;mp<totalMPs;mp++)
    {                
        int edgeIndx = 0;
        int vertsNum = 0;
        int i = I(mp);
        int j = J(mp);
        int k = K(mp);
        
        int mpFlag = (field[mp]<isoValue);
        int tempFlag;
        
        if(k+1<nz)
        {
        tempFlag  = mpFlag ^ (field[INDX(i, j, k+1)]<isoValue);
        vertsNum  += tempFlag;
        edgeIndx += tempFlag*2;
        }
        
        if(j+1<ny)
        {
        tempFlag  = mpFlag ^ (field[INDX(i, j+1, k)]<isoValue);
        vertsNum  += tempFlag;
        edgeIndx += tempFlag*4;
        }
        
        if(j+1<ny && k+1<nz)
        {
        tempFlag  = mpFlag ^ (field[INDX(i, j+1, k+1)]<isoValue);
        vertsNum  += tempFlag;
        edgeIndx += tempFlag*8;
        }
        
        if(i+1<nx)
        {
        tempFlag  = mpFlag ^ (field[INDX(i+1, j, k)]<isoValue);
        vertsNum  += tempFlag;
        edgeIndx += tempFlag*16;
        }
        
        if(i+1<nx&&k+1<nz)
        {
        tempFlag  = mpFlag ^ (field[INDX(i+1, j, k+1)]<isoValue);
        vertsNum  += tempFlag;
        edgeIndx += tempFlag*32;
        }
        
        if(i+1<nx&&j+1<ny)
        {        
        tempFlag  = mpFlag ^ (field[INDX(i+1, j+1, k)]<isoValue);
        vertsNum  += tempFlag;
        edgeIndx += tempFlag*64;
        }
        
        if(i+1<nx&&j+1<ny && k+1<nz)
        {        
        tempFlag  = mpFlag ^ (field[INDX(i+1, j+1, k+1)]<isoValue);
        vertsNum  += tempFlag;
        edgeIndx += tempFlag*128;
        }

        
        mpEdgeIndx[mp] = edgeIndx;
        mpVertsNum[mp] = vertsNum;
        mpActive[mp] = (vertsNum>0);
    }
    
    //Prefix-scan should be implemented in another way on GPU 
    
    
    //We apply a prefix-scan on mpVertsNum array
    //This will serve later as an address in verts
        
    mpVertsBase = (int*) malloc(sizeof(int)* totalMPs);
    mpVertsBase[0]=0;
    for(int mp=1;mp<totalMPs;mp++)
    {
        mpVertsBase[mp] = mpVertsBase[mp-1]+mpVertsNum[mp-1];
        
    }
    totalVerts = mpVertsBase[totalMPs-1]+mpVertsNum[totalMPs-1];
    printf("      Total Vertices: %d\n", totalVerts);
    
    //We build an array of active vertices and which enable us to process
    //mesh points that only has a non-zero number of associated crossed-edges.
    mpActiveScanned = (int*) malloc(sizeof(int)* totalMPs);
    mpActiveScanned[0]=0;
    for(int mp=1;mp<totalMPs;mp++)
    {
        mpActiveScanned[mp]=mpActiveScanned[mp-1]+mpActive[mp-1];
    }
    totalActive = mpActiveScanned[totalMPs-1]+mpActive[totalMPs-1];
    printf("      Total Active Mesh Points: %d\n", totalActive);
    
    if(totalActive==0)
    {
        free(mpActiveScanned);
        free(mpVertsNum);
        free(mpEdgeIndx);
        totalVerts = 0;
        totalTriangs = 0;
        return;
    }
    
    activeMPs =  (int*) malloc(sizeof(int)* totalActive);
    for(int mp=0;mp<totalMPs;mp++)
    {
        if(mpActive[mp])
        {
            activeMPs[mpActiveScanned[mp]]= mp;
        }
    }        

    
    free(mpActive);

    //1b-Calculate Voxel Normals:
    //----------------------------
    printf("    1b-Calculating Voxel Normals ...\n");
    voxNormals = (float*) malloc(sizeof(float)* 3 * totalMPs);
    for(int mp=0;mp<totalMPs;mp++)
    {
        int i = I(mp);
        int j = J(mp);
        int k = K(mp);

        float dfx,dfy,dfz, dx,dy,dz;
        
        int mp1,mp2;
        
        mp1 = INDX(i-1, j, k);
        mp2 = INDX(i+1, j, k);            
        if(i-1<0)
            mp1 = mp;
        if(i+1>=nx)
            mp2 = mp;            
        dfx = field[mp2] - field[mp1];
        dx = getX(mp2) - getX(mp1);
        
        
        mp1 = INDX(i, j-1, k);
        mp2 = INDX(i, j+1, k);            
        if(j-1<0)
            mp1 = mp;
        if(j+1>=ny)
            mp2 = mp;         
        dfy = field[mp2] - field[mp1];
        dy = getY(mp2) - getY(mp1);
        
        mp1 = INDX(i, j, k-1);
        mp2 = INDX(i, j, k+1);            
        if(k-1<0)
            mp1 = mp;
        if(k+1>=nz)
            mp2 = mp;         
        dfz = field[mp2] - field[mp1];
        dz = getZ(mp2) - getZ(mp1);

        voxNormals[mp*3] = dfx/dx;
        voxNormals[mp*3+1] = dfy/dy;
        voxNormals[mp*3+2] = dfz/dz;
    }
    
  //1c-Calculating Vertices & Normals:
  //-----------------------------------
    printf("    1c-Calculating Vertices Coordinates & Normals ...\n");
    
    verts = (float*) malloc(sizeof(float)* 3 * totalVerts);
    normals = (float*) malloc(sizeof(float)* 3 * totalVerts);
    
    for(int amp=0;amp<totalActive;amp++)
    {
        int mp0 = activeMPs[amp];
        int i0 = I(mp0);
        int j0 = J(mp0);
        int k0 = K(mp0);
        
        float f0 = field[mp0];
        float x0 = getX(mp0);
        float y_0 = getY(mp0);
        float z0 = getZ(mp0);
        
        float nx0 = voxNormals[mp0*3];
        float ny_0 = voxNormals[mp0*3+1];
        float nz0 = voxNormals[mp0*3+2];
        
        int edgeIndx = mpEdgeIndx[mp0];
        int vertsNum = mpVertsNum[mp0];
        
        for(int v=0;v<vertsNum;v++)
        {
            int edgeId = vert2EdgeId[v][edgeIndx/2];
            int i1 = i0+EDGE_DI(edgeId);
            int j1 = j0+EDGE_DJ(edgeId);
            int k1 = k0+EDGE_DK(edgeId);

            int mp1 = INDX(i1,j1,k1);
            
            float f1 = field[mp1];
            float x1 = getX(mp1);
            float y1 = getY(mp1);
            float z1 = getZ(mp1);
            

            float nx1 = voxNormals[mp1*3];
            float ny1 = voxNormals[mp1*3+1];
            float nz1 = voxNormals[mp1*3+2];
            
            float alpha = (isoValue - f0) / (f1 - f0);

            int vAddress = mpVertsBase[mp0]+v;
            float x =  x0 + alpha * (x1-x0);
            float y = y_0 + alpha * (y1-y_0);
            float z = z0 + alpha * (z1-z0);  
            verts[vAddress*3] = x;
            verts[vAddress*3+1] =y;
            verts[vAddress*3+2] = z;
            
            float nx =  nx0 + alpha * (nx1-nx0);
            float ny =  ny_0 + alpha * (ny1-ny_0);
            float nz =  nz0 + alpha * (nz1-nz0);
            float norm = sqrt(nx*nx+ny*ny+nz*nz);
            normals[vAddress*3] = nx/norm;
            normals[vAddress*3+1] = ny/norm;
            normals[vAddress*3+2] = nz/norm;
        }
    }    
    
    free(mpVertsNum);
    free(voxNormals);
    free(mpActiveScanned);

//2-Generate Triangles:
//======================
    printf("  2-Processing Triangles ...\n");
    
  //2a-Counting Triangles:
  //----------------------
    printf("    2a-Counting Triangles ...\n");
    voxTriangsNum = (int*) malloc(sizeof(int)* totalActive);
    for(int amp=0;amp<totalActive;amp++)
    {
        int mp = activeMPs[amp];
        int i = I(mp);
        int j = J(mp);
        int k = K(mp);

        if(!(i<nx-1&&j<ny-1&&k<nz-1))
        {
            voxTriangsNum[amp] = 0;
            continue;
        }
        
        int voxMP_Flags[8];
        voxMP_Flags[0] = (field[INDX(i, j, k)] < isoValue);
        voxMP_Flags[1] = (field[INDX(i, j, k+1)] < isoValue);
        voxMP_Flags[2] = (field[INDX(i, j+1, k)] < isoValue);
        voxMP_Flags[3] = (field[INDX(i, j+1, k+1)] < isoValue);
        voxMP_Flags[4] = (field[INDX(i+1, j, k)] < isoValue);
        voxMP_Flags[5] = (field[INDX(i+1, j, k+1)] < isoValue);
        voxMP_Flags[6] = (field[INDX(i+1, j+1, k)] < isoValue);
        voxMP_Flags[7] = (field[INDX(i+1, j+1, k+1)] < isoValue);          

        int vTriagsNum = 0;
        
        for(int tetra=0;tetra<6;tetra++)
        {
            int tetraIndx = 0;
            int bit = 1;
            for(int l=0;l<4;l++)
            {
                tetraIndx+=voxMP_Flags[tetraMP2voxMP[tetra][l]]*bit;
                bit*=2;
            }
            vTriagsNum += tetraTriangsNum[tetraIndx];
        }
        
        voxTriangsNum[amp] = vTriagsNum;
    }
    
    voxTriangsBase = (int*) malloc(sizeof(int)* totalActive);
    voxTriangsBase[0]=0;
    for(int amp=1;amp<totalActive;amp++)
    {
        voxTriangsBase[amp]=voxTriangsBase[amp-1]+voxTriangsNum[amp-1];
    }
    
    totalTriangs = voxTriangsBase[totalActive-1]+voxTriangsNum[totalActive-1];
    printf("      Total Triangles: %d\n", totalTriangs);
    free(voxTriangsNum);
  //2b-Calculating Triangles:
  //-------------------------    
    printf("    2b-Calculating Triangles ...\n");
    triangs = (GLuint*) malloc(sizeof(GLuint)* 3 * totalTriangs);
    for(int amp=0;amp<totalActive;amp++)
    {
        int mp = activeMPs[amp];
        int i = I(mp);
        int j = J(mp);
        int k = K(mp);
        if(!(i<nx-1&&j<ny-1&&k<nz-1))
        {
            continue;
        }
        int voxMPs[8];
        voxMPs[0] = mp;
        voxMPs[1] = INDX(i, j, k+1);
        voxMPs[2] = INDX(i, j+1, k);
        voxMPs[3] = INDX(i, j+1, k+1);
        voxMPs[4] = INDX(i+1, j, k);
        voxMPs[5] = INDX(i+1, j, k+1);
        voxMPs[6] = INDX(i+1, j+1, k);
        voxMPs[7] = INDX(i+1, j+1, k+1);
        
        int voxMP_Flags[8];
        for(int vmp=0;vmp<8;vmp++)
        {
            voxMP_Flags[vmp] = (field[voxMPs[vmp]] < isoValue);
        }
        

        int triangleShift = 0;
        for(int tetra=0;tetra<6;tetra++)
        {
            int tetraIndx = 0;
            int bit = 1;
            for(int tmp=0;tmp<4;tmp++)
            {
                tetraIndx+=voxMP_Flags[tetraMP2voxMP[tetra][tmp]]*bit;
                bit*=2;
            }
            int tetraVerts[6];
            int v = 0;
            for(int tmp1=0;tmp1<4;tmp1++)
            {
                for(int tmp2=tmp1+1;tmp2<4;tmp2++)
                {
                    int min, max;
                    if(tetraMP2voxMP[tetra][tmp1]>tetraMP2voxMP[tetra][tmp2])
                    {
                        min = tetraMP2voxMP[tetra][tmp2];
                        max = tetraMP2voxMP[tetra][tmp1];
                    }
                    else
                    {
                        max = tetraMP2voxMP[tetra][tmp2];
                        min = tetraMP2voxMP[tetra][tmp1];
                    }
                    int associatedMP = voxMPs[min];
                    int base = mpVertsBase[associatedMP];
                    
                    
                    int dI = EDGE_DI(max)-EDGE_DI(min);
                    int dJ = EDGE_DJ(max)-EDGE_DJ(min);
                    int dK = EDGE_DK(max)-EDGE_DK(min);
                    
                    int edgeId = EDGE_ID(dI,dJ,dK);
                    
                    
                    int shift = edge2VertShift[edgeId-1][mpEdgeIndx[associatedMP]/2];
                    if(shift>=0)
                        tetraVerts[v] = base + shift;
                    else
                        tetraVerts[v] = -1;
                    v++;
                }
            }
            
            for(int triangle=0;triangle<2;triangle++)
            {
                if(tetraTriangVerts[tetraIndx][3*triangle] < 0)
                        break;
                for(int vert = 0; vert < 3; vert++)
                {
                    int triangAddress = voxTriangsBase[amp]+triangleShift;

                    triangs[triangAddress*3+vert] =  
                    tetraVerts[tetraTriangVerts[tetraIndx][3*triangle+vert]];
                }
                triangleShift++;
            }            
        }
    }
    free(voxTriangsBase);
    
    free(mpEdgeIndx);
    free(activeMPs);
    free(mpVertsBase);
    printf("  Done!\n\n");

}

int main(int argc, char** argv)
{
    if(argc<3)
    {
        fprintf(stderr, 
                "Please specify field datafile followed by isovalue.\n");
        exit(1);
    }
    
    char* fileName = argv[1];
    isoValue = atof(argv[2]);
    
    loadField(fileName);
        
    generateRedBlueIso();    

    
    pi = atan(1)*4;
    r = 60;
    theta = pi / 2.5;
    phi = pi / 4.0;
    

    glutInit(&argc, argv);
    glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);
    glutInitWindowSize(800, 600);
    glutInitWindowPosition(0, 0);
    glutCreateWindow("Iso Surface");

    glutDisplayFunc(display);
    glutIdleFunc(display);    
    glutReshapeFunc(reshape);
    glutKeyboardFunc(processNormalKeys);
    glutMouseFunc(mouseButton);
    glutMotionFunc(mouseMove);

    glewInit();
    if (!glewIsSupported("GL_VERSION_2_0"))
    {
        printf("OpenGL 2.0 not supported\n");
        exit(1);
    }

    init();
    
    char* control = textFileRead("control.txt");
    printf("%s", control);
    free(control);

    glutMainLoop();
    return 0;
}

